import mwparserfromhell
import pywikibot
import re
from tugbot import lookup_dicts
from tugbot import free_text_parser
from tugbot import date_parser
from tugbot import complex_date_parser
from tugbot import dimensions_parser
from tugbot import medium_parser
from tugbot import provenance_parser
from tugbot import institution_parser
from tugbot import reference_parser


#===========================
site_wikidata = pywikibot.Site("wikidata", "wikidata")
repo_wikidata = site_wikidata.data_repository()

site_commons  = pywikibot.Site("commons", "commons")

#===========================
def is_category_definition_object(wikicode):
    wikicode = wikicode.lower()
    if("category definition: object")   in wikicode: return True
    elif("category_definition:_object") in wikicode: return True
    else: return False

def get_case_variations(string):
    l_words        = string.split()
    u_words        = string.title().split()
    combinations = [ l_words[0], u_words[0] ]
    for i in range(1, len(l_words)):
        tmp_vect = []
        for c in combinations:
            tmp_vect.append(c+" "+l_words[i])
            tmp_vect.append(c+" "+u_words[i])
        combinations = tmp_vect
    #print ("   ~~~", combinations)
    return combinations
    #return []

def get_q_number(item_label, lookup_dict=None):
    # Gets or guesses the Q-number relevant for 
    # a given label.
    if not item_label: return None
    item_label = item_label.lower().strip()
    
    # Easiest: label has hard-coded q-number in lookup dictionary
    if lookup_dict and (item_label in lookup_dict):
        return lookup_dict[item_label]
        
    # Now we retrieve Q-numbers by interrogating Wikidata.
    # TODO case-insensitive request
    # - If only one Q-number is return, we have a univoque item that
    #   we can safely return (e.g.: "Section Instruments de navigation")
    # - else we guess from the context (TODO)
    from pywikibot import pagegenerators
    
    entities = []
    if len(item_label.split(" ")) < 6:
        for i in get_case_variations(item_label):
            #print ("    querying", i)
            sparql = "SELECT ?item WHERE { ?item rdfs:label '"+i+"'@en }"
            entities = entities + list( pagegenerators.WikidataSPARQLPageGenerator(sparql, site=repo_wikidata) )
    else:
        sparql = "SELECT ?item WHERE { ?item rdfs:label '"+item_label+"'@en }"
        entities = pagegenerators.WikidataSPARQLPageGenerator(sparql, site=repo_wikidata)
        entities = list(entities)    
    
    # TODO we came up empty, looking in alt-labels   
    #if len(entities) == 0:
        #sparql = "SELECT ?item WHERE { ?item rdfs:altLabel '"+label+"'@en }"
        #entities = pagegenerators.WikidataSPARQLPageGenerator(sparql, site=repo_wikidata)
        #entities = list(entities)
    #print("  ~~~", entities)

    if len(entities) == 1:  # this means "if meaning of the label is unequivoqual"
        return entities[0].getID()
    
def get_accession_number(acc_text):
    # Translate "{{Louvre number|Sb 5793}}" into "Sb 5793"
    if re.search( r'[Ll]ouvre[ _][Nn]umber', acc_text):
        acc_text = acc_text.replace("}}", "")
        acc_text = acc_text.split("|")[1]
    #TODO See if there are other such templates and unwrap the data as needed
    return acc_text

#--------------------------
# Creates a Claim for one single date. Used to set dates or their qualifiers ("latest date", etc.)
def get_claim_from_date(data, P_date = lookup_dicts._P_date):
    claim = None
    
    # Year should always be set
    claim_target = str(data['year'])
    if 'month' in data:
        if data['month'] is not None:
            claim_target = claim_target + "-" + str(data['month'])
            if 'day' in data:
                if data['day'] is not None:
                    claim_target = claim_target + "-" + str(data['day'])

    # Sets precision
    claim_precision = 9 # default is Year
    if 'precision' in data:
        if data['precision'] is not None: claim_precision = lookup_dicts._epoch_dict[data['precision']]
    
    claim = pywikibot.Claim(repo_wikidata, P_date)
    claim.setTarget( pywikibot.WbTime( claim_target, precision=claim_precision ) )   #give a WbTime as argument
    
    return claim
#--------------------------
def is_circa(data):
    if 'adj' in data:
        if data['adj'] == 'circa': 
            return True
    return False


#===========================
class Cat_def_parser:
        
    def __init__(self, wikicode):
        wikicode_parsed = mwparserfromhell.parse(wikicode)
        templates = wikicode_parsed.filter_templates()
    
        self.fields   = {}
        
        for template in templates:
            if template.name.matches("Category definition: Object"):
                for i in template.params:
                    if i.split("="):
                        key   = i.split("=", 1)[0].strip()
                        value = i.split("=", 1)[1].strip()
                        if value and value != "":
                            self.fields[key] = value
                         
        # Spectific parsers for potentially complicated objects
        self.institution = institution_parser.Institution_parser(self.get_field("institution"))
        self.references  = reference_parser.Reference_parser(self.get_field("references"))

        
                    
    def get_field(self, field_name):
        if field_name in self.fields:
            return self.fields[field_name]
        return None
    
    #------------------------
    @property
    def image(self):
        return self.get_field("image")
    
    @property
    def type(self):
        return self.get_field("type")
    
    @property
    def artist(self):
        return self.get_field("artist")
    
    @property
    def location(self):    
        return self.get_field("location")
    
    @property
    def notes(self):    
        return self.get_field("notes")
    
    @property
    def title(self):
        title_field_text = self.get_field("title")
        return title_field_text

    @property
    def accession_number(self):    
        return self.get_field("accession number")    
    
    def print_out_free_text(self, field):
        free_text = free_text_parser.Free_text_parser(self.get_field(field))
        for lang in free_text.languages:
            print ( "         " + lang +"\t"+ free_text.get_text(lang) )

    def get_printable_free_text(self, field):
        work = ''
        free_text = free_text_parser.Free_text_parser(self.get_field(field))
        for lang in free_text.languages:
            work = work + "\n         " + lang +"\t"+ free_text.get_text(lang) 
        return work

    def get_free_text_data(self, field):
        free_text = free_text_parser.Free_text_parser(self.get_field(field))
        return free_text.get_data()
    
    def get_date_data(self):
        date_text = self.get_field("date")
        if complex_date_parser.is_complex_date(date_text):
            date = complex_date_parser.Complex_date_parser(date_text)
            return date.dates
        date = date_parser.Date_parser(date_text)
        return date.get_data()
    
    def get_size_data(self):
        size_text = self.get_field("dimensions")
        if dimensions_parser.Dimensions_parser(size_text):
            size = dimensions_parser.Dimensions_parser(size_text)
            return size.get_data() 
        #date = date_parser.Date_parser(size_text)
        #return date.get_data()  
        
    def get_medium_data(self):
        medium_text = self.get_field("medium")
        if medium_parser.Medium_parser(medium_text):
            medium = medium_parser.Medium_parser(medium_text)
            return medium.get_data() 

    def get_medium_q_numbers(self):
        medium_text = self.get_field("medium")
        if medium_parser.Medium_parser(medium_text):
            medium = medium_parser.Medium_parser(medium_text)
            return medium.get_q_numbers() 

        
    def get_provenance_data(self):
        provenance_text = self.get_field("object history")
        if provenance_parser.Provenance_parser(provenance_text):
            provenance = provenance_parser.Provenance_parser(provenance_text)
            #print("   ~~~~~", provenance.type, provenance.type_q_number )
            return provenance.get_data()
        
    def print_out_provenance_data(self):
        provenance_text = self.get_field("object history")
        if provenance_parser.Provenance_parser(provenance_text):
            provenance = provenance_parser.Provenance_parser(provenance_text)
            if provenance.type: 
                if provenance.type_q_number: print("    Type     : "+provenance.type+"\t("+provenance.type_q_number+")")
                else: print("    Type     : "+provenance.type+"\t(None)")
            if provenance.date: print("    Date     : "+provenance.date)
            if provenance.old_owner: print("    Old owner: "+provenance.old_owner+"\t("+str(get_q_number(provenance.old_owner, lookup_dicts._artist_Q_numbers))+")"   )        
            if provenance.new_owner: print("    New owner: "+provenance.new_owner+"\t("+str(get_q_number(provenance.new_owner, lookup_dicts._artist_Q_numbers))+")"   )


    #------------------------   
        
    @property
    def type_q_number(self):
        return get_q_number(self.type, lookup_dicts._type_Q_numbers)

    @property
    def location_q_number(self):
        return get_q_number(self.location)
    
    @property
    def artist_q_number(self):
        # If the artist is unknown:
        if not self.artist: return None
        if "anonymous" in self.artist.lower(): return "Q4233718"
        if "unknown"   in self.artist.lower(): return "Q4233718"
        
        # If the "artist" field has a {{Creator:}} template, retrieve the Q-Number from there
        if ("{{Creator:" in self.artist):
            creator_page_title = self.artist.replace("{{", "")
            creator_page_title = creator_page_title.replace("}}", "")
            creator_page_text  = pywikibot.Page(site_commons, creator_page_title).text
            if ("#REDIRECT" in creator_page_text):  # double redicts should not happen
                creator_page_title = creator_page_text.replace("#REDIRECT[[", "")
                creator_page_title = creator_page_title.replace("]]", "")
                creator_page_text  = pywikibot.Page(site_commons, creator_page_title).text
            matchObj = re.search( r'[wW]ikidata *= *(Q\d+)', creator_page_text)
            if matchObj: return matchObj.group(1)
        
        # Else we perform a standard search using the label as a feature
        return get_q_number(self.artist, lookup_dicts._artist_Q_numbers)
        
    #---------------------------
    def __str__(self):
        work = "  Image           :\t" + self.image
        work = work + "\n  Type            :\t"+self.type + "\t" + str(self.type_q_number)
        work = work + "\n  Artist          :\t"+ self.artist + "\t" + str(self.artist_q_number)
        work = work + "\n  Title           :" 
        work = work + self.get_printable_free_text("title")
        work = work + "\n  Description     :" 
        work = work + self.get_printable_free_text("description")
        work = work + "\n  Date            :\t"+ str(self.get_date_data() )
        work = work + "\n  Dimensions      :\t" + str( self.get_size_data() )
        work = work + "\n  Medium          :\t" + str( self.get_medium_data() ) + "\t" + str( self.get_medium_q_numbers() )
        work = work + "\n  Inscriptions    :" 
        work = work + self.get_printable_free_text("inscriptions")
        work = work + "\n  Object history  :\t" + str( self.get_provenance_data() )
        work = work + "\n  Institution     :\t" + self.institution.get_data() + "\t" + self.institution.get_q_number() 
        work = work + "\n  Location        :\t"+ str( self.location ) + "\t" + str(self.location_q_number)
        work = work + "\n  Accession number:\t"+get_accession_number(self.accession_number)
        work = work + "\n  Notes           :" 
        work = work + self.get_printable_free_text("notes")
        work = work + "\n  References      :"
        work = work + str( self.references )

        return work
                    
    
    #---------------------------
    # Methods to send information to Wikidata
    from pywikibot import FilePage
    #site_wikidata = pywikibot.Site("wikidata", "wikidata")
    site_wikidata = pywikibot.Site("test", "wikidata")
    repo_wikidata = site_wikidata.data_repository()
    #.....
    def create_wikidata_item(self, labels_dict):
        self.labels_dict = labels_dict
        self.wikidata_item_id = "Q4115189"
        #self.wikidata_item = pywikibot.ItemPage(site_wikidata)
        #self.wikidata_item.editLabels(labels=labels_dict, summary="Setting labels")
        #self.wikidata_item_id = self.wikidata_item.getID()
        self.wikidata_item    = pywikibot.ItemPage(repo_wikidata, self.wikidata_item_id)
    
    def commit_image(self):
        claim        = pywikibot.Claim(repo_wikidata, lookup_dicts._P_image)
        claim_target = pywikibot.FilePage(site_commons,  self.image )
        claim.setTarget( claim_target  )   #give a Commons page as argument !
        self.wikidata_item.addClaim(claim, summary=u'Tugbot setting Image property')
        
    def commit_property_by_Q_number(self, P_number, Q_number):
        if Q_number is None: return None
        claim        = pywikibot.Claim(repo_wikidata, P_number)
        claim_target = pywikibot.ItemPage(repo_wikidata, Q_number)
        claim.setTarget( claim_target  )
        self.wikidata_item.addClaim(claim, summary=u'Tugbot setting a generic property')
    
    def commit_property_by_free_text(self, P_number, free_texts_dict):
        claims = []
        for lang in free_texts_dict:
            claim        = pywikibot.Claim(repo_wikidata, P_number)
            claim.setTarget ( pywikibot.WbMonolingualText( free_texts_dict[lang], lang ) )
            claims.append(claim)
        if len(claims) == 0: return None
        for claim in claims:
            self.wikidata_item.addClaim(claim, summary=u'Tugbot setting a free-form property')
            
    def commit_property_by_string(self, P_number, string):            
        claim = pywikibot.Claim(repo_wikidata, P_number)
        claim.setTarget( string )   #give a string as argument
        self.wikidata_item.addClaim(claim, summary=u'Tugbot setting a free-form property')
            
    def commit_description(self):
        descriptions = self.get_free_text_data("description")
        if len (descriptions) == 0: return None
        self.wikidata_item.editDescriptions(descriptions, summary=u'Tugbot setting descriptions')

    def commit_date(self):
        data = self.get_date_data()
        claim0 = None
        claim0_is_circa = False
        
        if 'date0' not in data:
            claim0 = get_claim_from_date(data, lookup_dicts._P_date)
            if is_circa(data): claim0_is_circa = True
        else:
            claim0 = get_claim_from_date(data['date0'], lookup_dicts._P_date)
            if is_circa(data['date0']): claim0_is_circa = True

        if claim0: 
            self.wikidata_item.addClaim(claim0, summary='Tugbot setting Inception property')
            if claim0_is_circa:
                qualifier = pywikibot.Claim(repo_wikidata, lookup_dicts._P_circa)
                qualifier_target = pywikibot.ItemPage(repo_wikidata, lookup_dicts._Q_circa)
                qualifier.setTarget(qualifier_target)
                claim0.addQualifier(qualifier, summary=u'Tugbot adding qualifier Circa.')
            if 'date1' in data:
                claim1 = get_claim_from_date(data['date1'], lookup_dicts._P_earliest_date)
                claim0.addQualifier(claim1, summary='Tugbot adding earliest date.')
            if 'date2' in data:
                claim2 = get_claim_from_date(data['date2'], lookup_dicts._P_latest_date)
                claim0.addQualifier(claim2, summary='Tugbot adding latest date.')
    
    def commit_commons_category(self):
        commons_category_name = self.labels_dict['en']
        claim = pywikibot.Claim(repo_wikidata, lookup_dicts._P_commons_cat)
        claim.setTarget( commons_category_name )   #give a string as argument
        self.wikidata_item.addClaim(claim, summary='Tugbot setting Commons Category property')

    def commit_dimensions(self):
        length_dims = {
                        "height"   : lookup_dicts._P_heigth,
                        "width"    : lookup_dicts._P_width, 
                        "depth"    : lookup_dicts._P_thickness,
                        "length"   : lookup_dicts._P_length,
                        "diameter" : lookup_dicts._P_diameter,
                        "thickness": lookup_dicts._P_thickness
                        }
        mass_dims   = {
                        "mass": lookup_dicts._P_mass
                        }
        adim_dims   = {
                        "scale": lookup_dicts._P_scale
                        }
        
        dims = self.get_size_data()
        for dim in dims:
            claim = None
            if dim in length_dims.keys():
                claim = pywikibot.Claim(repo_wikidata, length_dims[dim])
                unit = pywikibot.ItemPage(repo_wikidata, lookup_dicts._unit_Q_numbers[ str(dims['length_unit']) ])
                claim.setTarget( pywikibot.WbQuantity( amount=dims[dim], unit=unit, error=0, site=repo_wikidata ) )
            elif dim in mass_dims.keys():
                claim = pywikibot.Claim(repo_wikidata, mass_dims[dim])
                unit = pywikibot.ItemPage(repo_wikidata, lookup_dicts._unit_Q_numbers[ str(dims['mass_unit']) ])
                claim.setTarget( pywikibot.WbQuantity( amount=dims[dim], unit=unit, error=0, site=repo_wikidata ) )
            elif dim in adim_dims.keys():
                claim = pywikibot.Claim(repo_wikidata, adim_dims[dim])
                claim.setTarget( pywikibot.WbQuantity( amount=dims[dim], error=0, site=repo_wikidata ) )
                
            if claim:
                self.wikidata_item.addClaim(claim, summary='Tugbot setting '+dim+' property')
    
    def commit_medium(self):
        for i in self.get_medium_q_numbers():
            self.commit_property_by_Q_number(lookup_dicts._P_medium, i)
            
    def commit_provenance(self):
        claim0        = pywikibot.Claim(repo_wikidata, lookup_dicts._P_institution)
        if not self.institution.get_q_number(): return None
        claim0_target = pywikibot.ItemPage(repo_wikidata, self.institution.get_q_number() )
        claim0.setTarget( claim0_target  )
        self.wikidata_item.addClaim(claim0, summary=u'Tugbot setting Collection')
        
        data = self.get_provenance_data()
        data_type_dict = {
                            'date'    : lookup_dicts._P_start_time,
                            'time'    : lookup_dicts._P_start_time,
                            'type'    : lookup_dicts._P_has_cause,
                            'newowner': lookup_dicts._P_donated_by,
                            'oldowner': lookup_dicts._P_donated_by
                        }
        for i in data:
            P_number = data_type_dict[i]
            #print(P_number)
            claim = None 
            if (i == 'type'):
                claim = pywikibot.Claim(repo_wikidata, P_number)
                Q_number =  lookup_dicts._provenance_Q_numbers[ str(data[i]) ]
                claim_target = pywikibot.ItemPage(repo_wikidata, Q_number)
                claim.setTarget( claim_target )
            if (i == 'date') or (i == 'time'):
                date = date_parser.Date_parser(data[i]).get_data()
                claim = get_claim_from_date(date, P_number)
            if (i == 'newowner') or (i == 'oldowner'):
                Q_number = get_q_number(data[i], lookup_dict=lookup_dicts._artist_Q_numbers)
                if Q_number:
                    claim = pywikibot.Claim(repo_wikidata, P_number)
                    claim_target = pywikibot.ItemPage(repo_wikidata, Q_number)
                    claim.setTarget( claim_target )
            if claim:
                claim0.addQualifier(claim, summary='Tugbot adding provenance information.') 
                
#.....
    def commit(self, labels_dict):
        self.create_wikidata_item(labels_dict) # Does not actually create anything for now. Will edit this when completed.
        #self.commit_image()
        #self.commit_property_by_Q_number(lookup_dicts._P_type,   self.type_q_number  )
        #self.commit_property_by_Q_number(lookup_dicts._P_artist, self.artist_q_number)
        #self.commit_property_by_free_text(lookup_dicts._P_title, self.get_free_text_data("title"))
        #self.commit_description()
        #self.commit_date()
        #self.commit_commons_category()
        #self.commit_dimensions()
        #self.commit_medium()
        #self.commit_property_by_free_text(lookup_dicts._P_inscriptions, self.get_free_text_data("inscriptions"))
        #self.commit_provenance()   
        #self.commit_property_by_Q_number(lookup_dicts._P_location, self.location_q_number  )
        #self.commit_property_by_string( lookup_dicts._P_accession_number, get_accession_number(self.accession_number) )
        #self.commit_property_by_free_text(lookup_dicts._P_title, self.get_free_text_data("notes"))
        
        #TODO write references
        #TODO code auto-edit of the original Commons page.
