import mwparserfromhell
import re
import math
from tugbot import date_parser


#==========================================                                
def parse_complex_date(date_text):
    fields = {}
    wikicode_parsed = mwparserfromhell.parse(date_text)
    templates = wikicode_parsed.filter_templates()
        
    for template in templates:
        if template.name.matches("complex date"):
            for i in template.params:
                if i.split("="):
                    key   = i.split("=", 1)[0].strip()
                    value = None
                    if (len(i.split("="))>1): value = i.split("=", 1)[1].strip()
                    if value and value != "":
                        fields[key] = value
    return fields

def get_century(date):
    work = int(date.year)
    if work > 0:
        work = math.ceil(work/100)
    else: 
        work = math.floor(work/100)
    return str(work)

def get_millennium(date):
    work = int(date.year)
    if work > 0:
        work = math.ceil(work/1000)
    else: 
        work = math.floor(work/1000)
    return str(work)

def is_complex_date(wikicode):
    if not wikicode: return False
    #if ( "{{complex date" in wikicode ) or ( "{{complex_date" in wikicode ): return True
    if re.search( r'{{[Cc]omplex[ _][Dd]ate', wikicode): return True
    return False

#==========================================                                

#------------
class Complex_date_parser:
    # Complex date template with syntax
    #    {{Complex date |conj= |date1= |adj1= |precision1= |era1= |date2= |adj2= |precision2= |era2= }}
    def __init__(self, wikicode):
        self.fields = {}
        wikicode = wikicode.lower()
        
        #if ( "{{complex date" in wikicode ) or ( "{{complex_date" in wikicode ):
        if is_complex_date(wikicode):
            self.fields = parse_complex_date(wikicode)
            
            # Common error in Complex_date usage usually denoting the desire to use "date1"
            if "date" in self.fields: self.fields["date1"] = self.fields["date"]
                   
            for date, precision, era, adj in zip(["date1", "date2"], ["precision1", "precision2"], ["era1", "era2"], ["adj1", "adj2"]):
                if date in self.fields: # If this is not the case for date1, we have a problem...
                    date_text = self.fields[date]
                    if precision in self.fields: date_text = date_text + " " + self.fields[precision]
                    if era       in self.fields: date_text = date_text + " " + self.fields[era]
                    if adj       in self.fields: date_text = date_text + " " + self.fields[adj]
                    #self.fields[date] = date_parser.Date_parser( date_text ).get_data()
                    self.fields[date] = date_parser.Date_parser( date_text )                    
                    
                    
        # The main date (self.fields["date0"]) is what we set as a Wikidata Property;
        # date1 and date2 are used as qualifiers for "earliest date" and "latest date"
        self.set_main_date()
        
#------------
    def get_data(self):
        return self.fields       
    
    @property
    def dates(self):
        dates = {}
        for date in ["date0", "date1", "date2"]:
            if date in self.fields: dates[date] = self.fields[date].get_data()
        if ( "date1" in self.fields ) and not ( "date2" in self.fields ): return self.fields["date0"].get_data()
        return dates
                        

    def set_main_date(self):
        # In case we have only one date, easy-peasy lemon squizy
        if ( "date1" in self.fields ) and not ( "date2" in self.fields ):
            self.fields["date0"] = self.fields["date1"]
        #elif ( "date1" in self.fields ) and ( "date2" in self.fields ) and (self.fields["date1"] == self.fields["date2"]):
            #self.fields["date0"] = self.fields["date1"]
            
        # In case we have two dates ("between"), then Wikidata will get these
        # two dates as "earliest date" and "latest dates" attributes to a 
        # main date that needs to be computed.
        elif ( "date1" in self.fields ) and ( "date2" in self.fields ):
            int_millennium1 = int(get_millennium(self.fields["date1"]))
            int_millennium2 = int(get_millennium(self.fields["date2"]))
            int_date1       = int( self.fields["date1"].year )
            int_date2       = int( self.fields["date2"].year )
            # if same century
            if ( get_century(self.fields["date1"]) == get_century(self.fields["date2"]) ):
                date_text = get_century(self.fields["date1"]) + " century"
                self.fields["date0"] = date_parser.Date_parser( date_text )
            # if same millennium
            elif ( get_millennium(self.fields["date1"]) == get_millennium(self.fields["date2"]) ):
                date_text = get_millennium(self.fields["date1"]) + " millennium"
                self.fields["date0"] = date_parser.Date_parser( date_text )
            # if neighouring millennia. If you have over two millenia difference, seriously fuck off.
            elif abs( int_millennium1 - int_millennium2 ) == 1:
                mean = ( int_date1 + int_date2 )/2
                date_text = get_millennium( date_parser.Date_parser( str(mean) ) )  + " millennium"
                self.fields["date0"] = date_parser.Date_parser( date_text )
                self.fields["adj0"] = "circa"
                
                


#TODO: Manage adjectives such as {{Complex date|mid|6|precision=century|era=BCE}}
# See https://commons.wikimedia.org/wiki/Category:Applique_with_character-Sb_3362













